#include "ball.h"
#include <float.h>

Ball::Ball() {
	cx=cy=vx=vy=ax=ay=0;
	radius = mass = 1;
	colID = -1;
}
void
Ball::setRadius(double r) {
	if (r > 0) {
		radius = r;
		mass = r*r*r;
	}
}
double
Ball::getRadius() const {
	return radius;
}
double
Ball::getX() const {
	return cx;
}
double
Ball::getY() const {
	return cy;
}
void
Ball::setPos(double x, double y) {
	cx = x;
	cy = y;
}
void
Ball::setSpeed(double x, double y) {
	vx = x;
	vy = y;
}
void
Ball::setAccel(double x, double y) {
	ax = x;
	ay = y;
}
unsigned int
Ball::getColor() const {
	return color;
}
/* the unsigned int stores a color value as
	r*2^16 + g*2^8 + b (+alpha*2^24)
   This is also the way in which Qt stores colors in uints. */
void
Ball::setColor(unsigned int c) {
	color = c;
}
void
Ball::setColor(int r, int g, int b) {
	color = (0xff << 24) |
		((r & 0xff) << 16) | ((g & 0xff) << 8) | (b & 0xff);
}
/* Normally the mass is deduced from the radius, but it may be overridden with
   this function. */
void
Ball::setMass(double m) {
	mass = m;
}
double
Ball::getXVel() const {
	return vx;
}
double
Ball::getYVel() const {
	return vy;
}
void
Ball::move(double t) {
	cx += t*vx + ax*t*t/2;
	cy += t*vy + ay*t*t/2;
	vx += t*ax;
	vy += t*ay;
	colID = -1;
}
double
Ball::getCollideTime(const Wall &w, int ID) const {
	double dist,t;
	t = -1;
	dist = 0;
	if (finite(w.getX())) {
		if (colID != ID) {
			dist = getOneDWallDistance(w.getX(), cx);
		}
		t = OneDCollideTime(dist, vx, ax);
		return t;
	} else if (finite(w.getY())) {
		if (colID != ID) {
			dist = getOneDWallDistance(w.getY(), cy);
		}
		t = OneDCollideTime(dist, vy, ay);
		return t;
	}
	return -1;
}
// This function calculates the distance between a wall and a ball.
// The result is exactly reproducable in different functions.
// This is important for a wellbehaved set of balls.
double
Ball::getOneDWallDistance(double wpos, double c) const {
	double d;
	d = wpos-c;
	if (d < -radius) {
		return d + radius;
	} else if (d > radius) {
		return d - radius;
	}
	return 0;
}
double
Ball::OneDCollideTime(double dist, double v, double a) const {
	double D, t1, t2;

	if (v == 0 && a == 0) {
		return -1;
	}
	if (a == 0) {
		return dist/v;
	}
	if (dist == 0) { // special solution, isolating it avoids num. err.
		return -v*2/a;
	}
	D = v*v+2*a*dist;
	if (D < 0) {
		return -1;
	}
	t1 = (-v+sqrt(D))/a;
	t2 = (-v-sqrt(D))/a;
	if (t1 > 0 && (t1 <= t2 || t2 < 0)) {
		return t1;
	} else if (t2 > 0 && (t2 <= t1 || t1 < 0)) {
		return t2;
	}
	return -1;
}
/* this function finds the time at which the two balls will collide
	and return -1 if the balls will never collide */
double
Ball::getCollideTime(const Ball &b, int ID) const {
	int i;
	double p[5], z[4];
	double dcx,dcy,dvx,dvy,dax,day;
	double t, t1, t2, ta, tb;
	double D;

	// Only the relative speed, relative position and relative acceleration
	// are relevant.
	dax = ax-b.ax;
	day = ay-b.ay;
	dvx = vx-b.vx;
	dvy = vy-b.vy;
	dcx = cx-b.cx;
	dcy = cy-b.cy;
	if (dax == 0 && day == 0) {
		// calculate time for balls with equal acceleration
		ta = dvx*dvx+dvy*dvy;
		if (ta == 0) { // if relative speed == 0 balls won't collide
			return -1;
		}
		tb = dcx*dvx+dcy*dvy;
		if (colID == ID) {
			return -2*tb/ta;
		}
		D = ta*pow(radius+b.radius,2) - pow(dvx*dcy-dvy*dcx,2);
		if (D < 0) {
			return -1;
		}
		if (D == 0) {
			return -tb/ta;
		}
		t1 = -(tb+sqrt(D))/ta;
		t2 = -(tb-sqrt(D))/ta;
		if (t1 > 0 && (t1 <= t2 || t2 < 0)) {
			return t1;
		} else if (t2 > 0 && (t2 <= t1 || t1 < 0)) {
			return t2;
		}
		return -1;
	}
	// simple solutions may be filtered out in the future
	// we must find all positive, real roots of a quartic (fourth-deg.)
	// polynomial p[4]*t^4 + p[3]*t^3 + p[2]*t^2 + p[1]*t + p[0] = 0
	p[4] = 0.25*(dax*dax+day*day);
	p[3] = dvx*dax + dvy*day;
	p[2] = dcx*dax + dcy*day + dvx*dvx + dvy*dvy;
	p[1] = 2*(dcx*dvx + dcy*dvy);
	if (colID == ID) {
		p[0] = 0;
		i = solveCubic(p+1,z);
	} else {
		p[0] = dcx*dcx + dcy*dcy - pow(radius+b.radius,2);
		i = solveQuartic(p,z);
	}
	t = -1;
	--i;
	while (i>=0) {
		if (z[i] > 0 && (t == -1 || z[i] < t)) {
			t = z[i];
		}
		--i;
	}
	return t;
}
void
Ball::collide(const Wall &w, int ID) {
	if (finite(w.getX())) {
		// make distance 0 to corr. num. errors
		cx -= getOneDWallDistance(w.getX(),cx);
		vx *= -1;
	} else if (finite(w.getY())) {
		// make distance 0 to corr. num. errors
		cy -= getOneDWallDistance(w.getY(),cy);
		vy *= -1;
	}
	colID = ID;
}
/* When we have arrived at a time where two balls will collide, this function
   must be called to calculate the new speeds for the balls. */
void
Ball::collide(Ball &b, int ID) {
	double d,dx,dy;
	double vp1, vl1, vp2, vl2;
	double newvp1, newvp2;

	colID = ID;

	// calculate unit vector along centra of balls
	dx = b.cx-cx;
	dy = b.cy-cy;
	if (dx == 0 && dy == 0 || mass == 0 || b.mass == 0) {
		return;
	}
	d = sqrt(dx*dx+dy*dy);
	dx /= d;
	dy /= d;

	// calculate parallel (p) and perpendicular (l) components
	vp1 = vx*dx+vy*dy;
	vl1 = vx*dy-vy*dx;
	vp2 = b.vx*dx+b.vy*dy;
	vl2 = b.vx*dy-b.vy*dx;

	// calculate new parallel speeds
	newvp1 = (vp1*(mass-b.mass) + vp2*2*b.mass)/(mass+b.mass);
	newvp2 = (vp1*2*mass + vp2*(b.mass-mass))/(mass+b.mass);

	// calculate new speeds
	vx = dx*newvp1+dy*vl1;
	vy = dy*newvp1-dx*vl1;
	b.vx = dx*newvp2+dy*vl2;
	b.vy = dy*newvp2-dx*vl2;
}
/* Return true if two balls touch or overlap. */
bool
Ball::overlap(const Ball &b) const {
	double r, dx, dy;
	r = radius + b.radius;
	dx = cx - b.cx;
	dy = cy - b.cy;
	return dx*dx + dy*dy <= r*r;
}
/* Return true if a ball and a wall touch or overlap. */
bool
Ball::overlap(const Wall &w) const {
	double d;
	if (finite(w.getX())) {
		d = cx-w.getX();
	} else {
		d = cy-w.getY();
	}
	return d*d <= radius*radius;
}
double
Ball::getKinEnergy() const {
	double e;
	// kinetic energy
	e = (vx*vx+vy*vy)*mass/2;
	return e;
}
double
Ball::getPotEnergy() const {
	double e=0;
	// potential energy
	e -= cx*mass*ax;
	e -= cy*mass*ay;
	return e;
}
