#include "bengcanvas.h"
#include "../bengbox.h"

BengBall::BengBall(double r, QCanvas *c, Ball *b)
	: QCanvasEllipse(int(2*r),int(2*r),c) {
	this->b = b;
	b->setRadius(r);
}
void
BengBall::setRadius(double r) {
	b->setRadius(r);
	if (r<1) {
		setSize(2,2);
	} else {
		setSize(int(2*r), int(2*r));
	}
}
void
BengBall::move() {
	QCanvasEllipse::move(b->getX(), b->getY());
}
BengCanvas::BengCanvas()
	: QCanvas() {
	radius = 0.5;
	polydisp = 0.5;
	n = 5;
	QCanvas::resize(1, 1);
	box = new BengBox(n, 1, 1);
	setBackgroundColor(Qt::green);
	setDoubleBuffering(true);
	ready = false;
	acceleration = 250;
	pi = 2*acos(0);
	queens = 2;

	reset();
}
BengCanvas::~BengCanvas() {
	delete box;
	vector<BengBall *>::iterator i;
	for(i=balls.begin(); i != balls.end(); ++i) {
		delete *i;
	}
}
void
BengCanvas::setPolydispersity(double p) {
	// if polydisp = 0, all balls have equal size
	// if polydisp = 1, ball have size ranging from max/n to max
	polydisp = p;
	if (polydisp < 0) polydisp = 0;
	if (polydisp > 1) polydisp = 1;
	placeBalls();
}
void
BengCanvas::setRadius(double r) {
	radius = r;
	if (radius <= 0) radius = 0.01;
	if (radius > 1) radius = 1;
	placeBalls();
}
void
BengCanvas::placeBalls() {
	unsigned int i;
	double maxr, r;
	maxr = sqrt(width()*height()/n)/3*radius;
	for (i=0; i<n; ++i) {
		r = maxr-polydisp*(n-i-1)*maxr/n;
		balls[i]->setRadius(r);
		if (i >= n-queens) {
			balls[i]->b->setMass(10000); // override mass
		}
	}
	box->removeOverlap();
}
void
BengCanvas::setNumBalls(uint n) {
	this->n = n;
	reset();
}
double
BengCanvas::getPolydispersity() const {
	return polydisp;
}
double
BengCanvas::getRadius() const {
	return radius;
}
uint
BengCanvas::getNumBalls() const {
	return n;
}
void
BengCanvas::setQueens(uint i) {
	if (i > n) {
		i = n;
	}
	queens = i;
	reset();
}
uint
BengCanvas::getQueens() const {
	return queens;
}
void
BengCanvas::resize(int w, int h) {
	QCanvas::resize(w, h);
	reset();
}
void
BengCanvas::reset() {
	int w, h;
	double x, y, r;
	unsigned int i;
	BengBall *bb, *queen;

	w = width();
	h = height();

	delete box;
	box = new BengBox(n, width(), height());
	for (i=0; i<balls.size(); ++i) {
		delete balls[i];
	}
	balls.resize(n); 

//	int max = int(ceil(sqrt(n)));
	double maxr = sqrt(w*h/n)/3*radius;
	for (i=0; i<n; i++) {
		r = maxr-polydisp*(n-i-1)*maxr/n;
		bb = new BengBall(r, this, box->getBall(i));
//		x = (0.1+0.8*(i/max)/sqrt(n))*w;
//		y = (0.5+0.4*(i%max)/sqrt(n))*h;
		x = 0;
		y = 0;
		bb->b->setPos(x, y);
		if (i >= n-queens) {
			bb->setBrush(red);
			bb->b->setMass(1000000); // override mass
			queen = bb;
		} else {
			bb->setBrush(QColor(i*256/n,i*256/n,i*256/n));
		}
		balls[i] = bb;
	}
	box->removeOverlap();
	for (i=0; i<n; i++) {
		bb = balls[i];
	//	if (i >= n-queens) {
			double angle = random()*2*pi/RAND_MAX;
			x = w/5*cos(angle);
			y = w/5*sin(angle);
	//	} else {
	//		x = w/5;
	//		y = 0;
	//	}
		bb->b->setSpeed(x, y);
		x = 0;
		if (queens) {
			y = 0;
		} else {
			y = 250;
		}
		bb->b->setAccel(x, y);
		bb->move();
		bb->show();
	}
	setUpdatePeriod(40);
	ready = true;
}
void
BengCanvas::setUpdatePeriod(int ms) {
	updatePeriod = ms;
	QCanvas::setUpdatePeriod(ms);
}
void
BengCanvas::update() {
	vector<BengBall *>::iterator it;
	box->step(0.001*updatePeriod);
	for (it=balls.begin(); it != balls.end(); ++it) {
		(*it)->move();
	}
	if (ready) {
		checkEnergy();
	}
	if (queens) {
		swarmBalls();
	}
	QCanvas::update();
}
void
BengCanvas::checkEnergy() {
	//static int i=0;
	double pe = 0, ke=0;
	vector<BengBall *>::iterator it;
	for (it=balls.begin(); it != balls.end(); ++it) {
		pe += (*it)->b->getPotEnergy();
		ke += (*it)->b->getKinEnergy();
	}
	if (fabs(pe+ke - energy) > 1) {
	}
}
void
BengCanvas::swarmBalls() {
	unsigned int i, j;
	double dx, dy, vx, vy, r, da;
	double finalangle, angle;
	double anglestep = pi/12;
	Ball *q;
	Ball *b;
	for (i=0; i<n-queens; i++) {
		b = balls[i]->b;
		dx = b->getX();
		dy = b->getY();
		// find the closest queen
		q = 0;
		r = 0;
		for (j=0; j<queens; j++) {
			Ball *newq = balls.at(n-j-1)->b;
			double newr = hypot(dx-newq->getX(), dy-newq->getY());
			if (newr < r || q == 0) {
				r = newr;
				q = newq;
			}
		}
		// adjust the direction
		b = balls[i]->b;
		dx = q->getX() - dx;
		dy = q->getY() - dy;
		vx = b->getXVel();
		vy = b->getYVel();
		r = sqrt(vx*vx+vy*vy);
		finalangle = atan2(dx, dy); // gives -0 < angle < 2 pi
		angle = atan2(vx, vy);
		da = angle - finalangle;
		if (da > pi) da -= 2*pi;
		if (da < -pi) da += 2*pi;
		if (da > anglestep) {
			angle -= anglestep;
		} else if (da < -anglestep) {
			angle += anglestep;
		} else {
			angle = finalangle;
		}
		vx = r*sin(angle);
		vy = r*cos(angle);
		b->setSpeed(vx, vy);
	}
}
/*void
BengCanvas::swarmBalls() {
	double x, y, dx, dy, vx, vy, r, da;
	double finalangle, angle;
	double anglestep = pi/12;
	BengBall *b;
	b=balls.last();
	x = b->b->getX();
	y = b->b->getY();
	b = balls.prev();
	while (b) {
		dx = x-b->b->getX();
		dy = y-b->b->getY();
		vx = b->b->getXVel();
		vy = b->b->getYVel();
		r = sqrt(vx*vx+vy*vy);
		finalangle = atan2(dx, dy); // gives -0 < angle < 2 pi
		angle = atan2(vx, vy);
		da = angle - finalangle;
		if (da > pi) da -= 2*pi;
		if (da < -pi) da += 2*pi;
		if (da > anglestep) {
			angle -= anglestep;
		} else if (da < -anglestep) {
			angle += anglestep;
		} else {
			angle = finalangle;
		}
		vx = r*sin(angle);
		vy = r*cos(angle);
		b->b->setSpeed(vx, vy);
		b = balls.prev();
	}
}*/
