#include "bengsaver.h"
#include "bengsaver.moc"
#include "bengsetup.h"
#include "../swarmbox.h"
#include "../ball.h"
#ifdef HAVE_GL
#include "glbox.h"
#endif
#include <kconfig.h>
#include <kglobal.h>
#include <qlayout.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qcanvas.h>
#include <klocale.h>

// libkscreensaver interface
extern "C" {
	const char *kss_applicationName = "bengsaver.kss";
	const char *kss_description = I18N_NOOP( "BengSaver" );
	const char *kss_version = "2.2.0";

	KScreenSaver *kss_create( WId id ) {
		return new BengSaver( id );
	}

	QDialog *kss_setup() {
		return new BengSetup();
	}
}

#include <iostream>
using std::cout;
using std::endl;

BengSaver::BengSaver(WId id)
	: KScreenSaver(id) {
	qc = NULL;
	qcv = NULL;
#ifdef HAVE_GL
	gl = NULL;
#endif
	sb = new SwarmBox();
	readSettings();
	ready = false;
	threeD = !threeD;
	set3D(!threeD);
	setPolydispersity(polydispersity);
	setRadius(radius);
	setNumBalls(numberOfBalls);
	setQueens(numberOfQueens);
	setBigColor(bigcolor);
	setSmallColor(smallcolor);
	setQueenColor(queencolor);

	// events to the screen saver can be filtered out, so we need to
	// check every now and then to see if the widgetsize has been changed,
	// so we can change the widgetsize
	timer = new QTimer(this);
	timer->start( sb->getUpdatePeriod(), TRUE);
	connect(timer, SIGNAL(timeout()), this, SLOT(update()));
}
BengSaver::~BengSaver() {
	delete sb;
}
void
BengSaver::readSettings() {
	KConfig *config = KGlobal::config();
	config->setGroup("Settings");

	numberOfBalls = config->readNumEntry("Number of Balls", 5);
	polydispersity = config->readDoubleNumEntry("Polydispersity", 0.5);
	radius = config->readDoubleNumEntry("Radius", 0.5);
	numberOfQueens = config->readNumEntry("Number of Queens", 0);
	backcolor = config->readColorEntry("Background color", &green);
	bigcolor = config->readColorEntry("Big ball color", &white);
	smallcolor = config->readColorEntry("Small ball color", &black);
	queencolor = config->readColorEntry("Queen ball color", &red);
	threeD = config->readBoolEntry("3D", true);
}
void
BengSaver::update() {
	if (!ready || width() != sb->width() || height() != sb->height()) {
		ready = true;
		if (qcv) {
			qcv->resize(width(), height());
			qc->resize(width(), height());
#ifdef HAVE_GL
		} else {
			gl->resize(width(), height());
#endif
		}
		sb->resize(width(), height());
	}
	timer->start(sb->getUpdatePeriod(), TRUE);

	sb->update();
	if (qcv) {
		updateCanvas();
#ifdef HAVE_GL
	} else {
		updateOpenGL();
#endif
	}
}
void
BengSaver::updateCanvas() {
	uint i;
	uint nn = sb->getNumBalls();
	uint n = qc->allItems().size();
	// adjust number of balls on canvas
	if (n != nn) {
		if (nn < n) {
			QCanvasItemList list = qc->allItems();
			QCanvasItemList::Iterator it;
			i = 0;
			for (it = list.begin(); it != list.end(); ++it) {
				if (i++ >= nn) {
					delete *it;
				}
			}
		} else {
			QCanvasEllipse *c;
			for (i=n; i<nn; ++i) {
				c = new QCanvasEllipse(qc);
			}
		}
	}
	QCanvasItemList list = qc->allItems();
	QCanvasItemList::Iterator it;
	i = 0;
	for (it = list.begin(); it != list.end(); ++it) {
		Ball *b = sb->getBall(i++);
		if (*it) {
			(*it)->move(b->getX(), b->getY());
			QCanvasEllipse *qce = static_cast<QCanvasEllipse *>(*it);
			qce->setBrush(QColor(b->getColor()));
			qce->setSize(int(b->getRadius()*2), int(b->getRadius()*2));
			qce->show();
		}
	}
	qc->update();
}
#ifdef HAVE_GL
void
BengSaver::updateOpenGL() {
	uint n = sb->getNumBalls();
	if (gl->getNumBalls() != n) {
		gl->setNumBalls(n);
	}
	for (uint i=0; i<n; ++i) {
		Ball *b = sb->getBall(i);
		gl->setBallCoord(i, b->getX(), b->getY());
		gl->setBallRadius(i, b->getRadius());
		gl->setBallColor(i, b->getColor());
	}
	gl->updateGL();
}
#endif
void
BengSaver::setPolydispersity(double p) {
	sb->setPolydispersity(p);
}
void
BengSaver::setRadius(double r) {
	sb->setRadius(r);
}
void
BengSaver::setNumBalls(uint nn) {
	sb->setNumBalls(nn);
}
void
BengSaver::setQueens(uint i) {
	sb->setQueens(i);
}
void
BengSaver::setBackColor(const QColor &c) {
	backcolor = c;
	if (qc) {
		qc->setBackgroundColor(c);
#ifdef HAVE_GL
	} else {
		gl->setBackgroundColor(c);
#endif
	}
}
void
BengSaver::setBigColor(const QColor &c) {
	sb->setBigColor(c.rgb());
}
void
BengSaver::setSmallColor(const QColor &c) {
	sb->setSmallColor(c.rgb());
}
void
BengSaver::setQueenColor(const QColor &c) {
	sb->setQueenColor(c.rgb());
}
void
BengSaver::set3D(bool threeD) {
	if (threeD == this->threeD) {
		return;
	} 
	this->threeD = threeD;
#ifdef HAVE_GL
	if (threeD) {
		if (qc) {
			delete qc;
			qc = NULL;
			delete qcv;
			qcv = NULL;
		}
		gl = new GLBox(this);
		gl->resize(width(), height());
		gl->setBackgroundColor(backcolor);
		gl->show();
	} else {
		if (gl) {
			delete gl;
		}
#endif
		qc = new QCanvas();
		qc->setDoubleBuffering(true);
		qc->setBackgroundColor(backcolor);
		qc->resize(width(), height());
		qcv = new QCanvasView(this);
		qcv->setVScrollBarMode(QScrollView::AlwaysOff);
		qcv->setHScrollBarMode(QScrollView::AlwaysOff);
		qcv->setFrameShape(QFrame::NoFrame);
		qcv->setCanvas(qc);
		qcv->resize(width(), height());
		qcv->show();
#ifdef HAVE_GL
	}
#endif
}
