#include "swarmbox.h"
#include "bengbox.h"

SwarmBox::SwarmBox() {
	// start with some sane values
	radius = 0.5;
	polydisp = 0.5;
	acceleration = 250;
	n = 5;
	queens = 2;
	ready = false;
	updatePeriod = 40; // milliseconds

	box = new BengBox(n, 1, 1);
	pi = 2*acos(0);

	bigcolor = smallcolor = queencolor = 0;
}
SwarmBox::~SwarmBox() {
	delete box;
}
void
SwarmBox::setPolydispersity(double p) {
	// if polydisp = 0, all balls have equal size
	// if polydisp = 1, ball have size ranging from max/n to max
	polydisp = p;
	if (polydisp < 0) polydisp = 0;
	if (polydisp > 1) polydisp = 1;
	if (ready) {
		placeBalls();
	}
}
void
SwarmBox::setRadius(double r) {
	radius = r;
	if (radius <= 0) radius = 0.01;
	if (radius > 1) radius = 1;
	if (ready) {
		placeBalls();
	}
}
void
SwarmBox::placeBalls() {
	unsigned int i;
	double maxr, r;
	maxr = sqrt(w*h/n)/3*radius;
	for (i=0; i<n; ++i) {
		r = maxr-polydisp*(n-i-1)*maxr/n;
		box->getBall(i)->setRadius(r);
		if (i >= n-queens) {
			box->getBall(i)->setMass(10000); // override mass
		}
	}
	box->removeOverlap();
}
void
SwarmBox::setNumBalls(uint n) {
	if (this->n != n) {
		this->n = n;
		if (ready) {
			reset();
		}
	}
}
double
SwarmBox::getPolydispersity() const {
	return polydisp;
}
double
SwarmBox::getRadius() const {
	return radius;
}
uint
SwarmBox::getNumBalls() const {
	return n;
}
void
SwarmBox::setQueens(uint i) {
	if (i > n) {
		i = n;
	}
	queens = i;
	if (ready) {
		reset();
	}
}
uint
SwarmBox::getQueens() const {
	return queens;
}
void
SwarmBox::resize(int w, int h) {
	this->w = w;
	this->h = h;
	reset();
}
void
SwarmBox::reset() {
	double x, y, r;
	unsigned int i;
	Ball *bb;

	delete box;
	box = new BengBox(n, w, h);

	double maxr = sqrt(w*h/n)/3*radius;
	for (i=0; i<n; ++i) {
		r = maxr-polydisp*(n-i-1)*maxr/n;
		bb = box->getBall(i);
		bb->setRadius(r);
		x = 0; // impossible position, will be corrected by
		y = 0; // removeOverlap()
		bb->setPos(x, y);
		if (i >= n-queens) {
			// initialize a queen
			bb->setMass(1000000); // override mass
		}
	}
	ready = true;
	updateColors();
	box->removeOverlap();
	for (i=0; i<n; i++) {
		bb = box->getBall(i);
		double angle = random()*2*pi/RAND_MAX;
		x = w/5*cos(angle);
		y = w/5*sin(angle);
		if (!queens) {
			x /= 4;
			y = 0;
		}
		bb->setSpeed(x, y);
		x = 0;
		if (queens) {
			y = 0;
		} else {
			y = 250;
		}
		bb->setAccel(x, y);
	}
}
int
SwarmBox::getUpdatePeriod() const {
	return updatePeriod;
}
void
SwarmBox::setUpdatePeriod(int ms) {
	updatePeriod = ms;
}
Ball *
SwarmBox::getBall(uint i) const {
	return box->getBall(i);
}
void
SwarmBox::update() {
	box->step(0.001*updatePeriod);
//	if (ready) {
//		checkEnergy();
//	}
	if (queens) {
		swarmBalls();
	}
}
/* this function is a first attempt to control the speed at which the balls move
   the bengbox has a constant energy, so it is the best control parameter for
   controlling the speed of the balls. */
void
SwarmBox::checkEnergy() {
	Ball *b;
	double pe = 0, ke=0;
	for (uint i=0; i<n; ++i) {
		b = box->getBall(i);
		pe += b->getPotEnergy();
		ke += b->getKinEnergy();
	}
	if (fabs(pe+ke - energy) > 1) {
	}
}
/* This function lets the balls that are not a queen adjust the direction in
   which they are going to the position of the nearest queen. */
void
SwarmBox::swarmBalls() {
	unsigned int i, j;
	double dx, dy, vx, vy, r, da;
	double finalangle, angle;
	double anglestep = pi/12;
	Ball *q;
	Ball *b;
	for (i=0; i<n-queens; i++) {
		b = box->getBall(i);
		dx = b->getX();
		dy = b->getY();
		// find the closest queen
		q = 0;
		r = 0;
		for (j=0; j<queens; j++) {
			Ball *newq = box->getBall(n-j-1);
			double newr = hypot(dx-newq->getX(), dy-newq->getY());
			if (newr < r || q == 0) {
				r = newr;
				q = newq;
			}
		}
		// adjust the direction
		b = box->getBall(i);
		dx = q->getX() - dx;
		dy = q->getY() - dy;
		vx = b->getXVel();
		vy = b->getYVel();
		r = sqrt(vx*vx+vy*vy);
		finalangle = atan2(dx, dy); // gives -pi < angle < pi
		angle = atan2(vx, vy);
		da = angle - finalangle;
		if (da > pi) da -= 2*pi;
		if (da < -pi) da += 2*pi;
		if (da > anglestep) {
			angle -= anglestep;
		} else if (da < -anglestep) {
			angle += anglestep;
		} else {
			angle = finalangle;
		}
		vx = r*sin(angle);
		vy = r*cos(angle);
		b->setSpeed(vx, vy);
	}
}
void
SwarmBox::setBigColor(unsigned int c) {
	bigcolor = c;
	updateColors();
}
void
SwarmBox::setSmallColor(unsigned int c) {
	smallcolor = c;
	updateColors();
}
void
SwarmBox::setQueenColor(unsigned int c) {
	queencolor = c;
	updateColors();
}
void
SwarmBox::updateColors() {
	uint i;
	Ball *b;
	for (i=0; i<n && ready; ++i) {
		b = box->getBall(i);
		if (i >= n-queens) {
			b->setColor(queencolor);
		} else {
			if (n == queens+1) {
				b->setColor(bigcolor);
			} else {
				setColor(i/double(n-queens-1), b);
			}
		}
	}
}
void
SwarmBox::setColor(double f, Ball *ball) const {
	int s = int((smallcolor >> 16) & 0xff);
	int b = int((bigcolor >> 16) & 0xff);
	int r = int(s + (b-s)*f);
	s = int((smallcolor >> 8) & 0xff);
	b = int((bigcolor >> 8) & 0xff);
	int g = int(s + (b-s)*f);
	s = int(smallcolor & 0xff);
	b = int(bigcolor & 0xff);
	b = int(s + (b-s)*f);
	ball->setColor(r, g, b);
}
int
SwarmBox::width() const {
	return w;
}
int
SwarmBox::height() const {
	return h;
}
